# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2016-04-30 18:32:09.694914
from __future__ import unicode_literals
import datetime

version = "1.1"
githash = "829410b24435ee678dbff3bf0eede5e1af993c39"


major = 1
minor = 1
bugfix = 0

release = True
timestamp = datetime.datetime(2016, 4, 30, 18, 32, 9, 694914)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
